<?php

/**
 * A library that uses the cybertill api to make orders and sync stock
 * SOAP API docs: http://ct28790.c-pos.co.uk/current/docs/soap_documentation_v1_5/index.html
 */

class Cybertill {

	//The current loaded config (live/dev)
	protected $_config = null;

	//The current Soap Client
	protected $_client = null;

	//The current authorisation id returned from the client
	protected $_authId = null;

	//The current customer id, taken from config if avilable created from customer details
	protected $_customer = null;

/**
 * Make an order on the client's cybertill system.
 * @param  array 		$order An EvCheckout order array consisting of order info, order items, order totals and order data.
 * @return array|bool   Returns transaction array. Contains success/failure and transaction info/errors.
 */
	public function makeOrder($order) {
		if ($this->_setupClient()) {
			$this->_setupCustomer();

			if (!empty($this->_customer)) {
				$cybertillOrderDetails = $this->_getDetailsForOrder($order);

				$cybertillOrderItems = $this->_getItemsForOrder($order);

				$cybertillOrderDelivery = $this->_getDeliveryForOrder($order);

				$cybertillOrderPayments = $this->_getPaymentsForOrder($order);

				if (
					$cybertillOrderDetails['success'] === true &&
					$cybertillOrderItems['success'] === true &&
					$cybertillOrderDelivery['success'] === true &&
					$cybertillOrderPayments['success'] === true
				) {
					$transaction = $this->_client->transaction_add_giftcard(
						$cybertillOrderDetails['data'],
						$cybertillOrderItems['data'],
						$cybertillOrderDelivery['data'],
						$cybertillOrderPayments['data']
					);

					if ($transaction->success) {
						if (!empty($this->_config['despatch_user_id'])) {
							$dispatchOrder = [
								[
									'transactionId' => $transaction->transaction->id,
									'locationId' => $this->_config['location_id'],
									'consignmentRef' => $this->_config['order_details']['order_note'],
									'userId' => null
								]
							];

							$dispatched = $this->_client->transaction_despatch_orders(
								$this->_config['despatch_user_id'],
								$dispatchOrder
							);

							if ($dispatched->success) {
								return ['success' => true];
							} else {
								$dispatchedError = $dispatched->errors->item->error;

								if (Configure::check('EvCybertill.errorLogging.dispatchFailure') && Configure::read('EvCybertill.errorLogging.dispatchFailure') === true) {
									if (CakePlugin::loaded('EvErrbit')) {
										$e = new Exception(print_r($dispatchedError, true));
										Errbit::notify($e);
									}

									CakeLog::write('cybertillApi', print_r($dispatchedError, true));
									CakeLog::write(
										'cybertillApi',
										"\n Transaction Data Sent: \n" .
										print_r($cybertillOrderDetails['data'], true) . "\n" .
										print_r($cybertillOrderItems['data'], true) . "\n" .
										print_r($cybertillOrderDelivery['data'], true) . "\n" .
										print_r($cybertillOrderPayments['data'], true)
									);
								}

								return ['success' => false, 'message' => $dispatchedError, 'orderCreated' => true, 'adminNotification' => 'dispatchFailure'];
							}
						} else {
							return ['success' => false, 'message' => 'Dispatch user missing from config.', 'orderCreated' => true, 'adminNotification' => 'dispatchFailure'];
						}
					} else {
						if (Configure::check('EvCybertill.errorLogging.transactionFailure') && Configure::read('EvCybertill.errorLogging.transactionFailure') === true) {
							if (CakePlugin::loaded('EvErrbit')) {
								$e = new Exception(print_r($transaction, true));
								Errbit::notify($e);
							}

							CakeLog::write('cybertillApi', print_r($transaction, true));
							CakeLog::write(
								'cybertillApi',
								"\n Transaction Data Sent: \n" .
								print_r($cybertillOrderDetails['data'], true) . "\n" .
								print_r($cybertillOrderItems['data'], true) . "\n" .
								print_r($cybertillOrderDelivery['data'], true) . "\n" .
								print_r($cybertillOrderPayments['data'], true)
							);
						}

						return ['success' => false, 'message' => $transaction->errors->item->error, 'adminNotification' => 'transactionFailure'];
					}
				} else {
					$orderFailureMessage = '';

					if (!empty($cybertillOrderDetails['message'])) {
						$orderFailureMessage .= "\n" . $cybertillOrderDetails['message'];
					}

					if (!empty($cybertillOrderItems['message'])) {
						$orderFailureMessage .= "\n" . $cybertillOrderItems['message'];
					}

					if (!empty($cybertillOrderDelivery['message'])) {
						$orderFailureMessage .= "\n" . $cybertillOrderDelivery['message'];
					}

					if (!empty($cybertillOrderPayments['message'])) {
						$orderFailureMessage .= "\n" . $cybertillOrderPayments['message'];
					}

					return ['success' => false, 'message' => $orderFailureMessage, 'adminNotification' => 'orderSetupFailure'];
				}
			} else {
				return ['success' => false, 'message' => 'Customer details missing', 'adminNotification' => 'customerSetupFailure'];
			}
		} else {
			return ['success' => false, 'message' => 'Unable to establish client', 'adminNotification' => 'clientSetupFailure'];
		}

		return ['success' => false];
	}

/**
 * Sync the stock between the website and cybertill. It doesn't update cybertill information.
 * @return bool Returns true if the stock has attempted to be updated. False if no items were
 *                      found to be synchronised.
 */
	public function syncStock() {
		if ($this->_setupClient()) {
			if (!empty($this->_config['models'])) {
				$models = $this->_config['models'];
				if (!empty($models['item_model']) && !empty($models['stock_model'])) {
					$ItemModel = EvClassRegistry::init($models['item_model']);

					$items = $ItemModel->find('all');

					foreach ($items as $item) {
						if (!empty($item[$ItemModel->alias]['sku'])) {
							$this->syncItemStock($item[$ItemModel->alias]['sku'], $item, $ItemModel);
						}
					}

					return true;
				}
			}
		}

		return false;
	}

	public function syncItemStock($sku, $item, $ItemModel) {
		if ($this->_setupClient()) {
			$cybertillItem = $this->_getItem($sku);

			if (!empty($cybertillItem['item']->productOption->id)) {
				$stock = $this->_getItemStock($sku, $cybertillItem['item']->productOption->id);

				if (!empty($stock)) {
					$this->_updateStock($stock, $item, $ItemModel);
				}
			}
		}
	}

/**
 * Check the stock of a particular item by using the SKU provided.
 * @param  string   $sku The SKU to be used to find the item to return the stock of
 * @return int|bool      If an item and stock are found then stock will be returned otherwise false is returned
 */
	public function checkStock($sku) {
		if (empty($this->_client)) {
			$this->_setupClient();
		}

		if (!empty($this->_client)) {
			$productItem = $this->_getItem($sku);

			if ($productItem['success'] === true) {
				return ['success' => true, 'stock' => $this->_getItemStock($sku, $productItem['item']->productOption->id)];
			} else {
				return ['success' => false, 'message' => $productItem['message']];
			}
		}

		return ['success' => false, 'message' => 'unable to establish client'];
	}

/**
 * Set the stock of an item by using the SKU provided.
 * @param string $sku         The SKU of the cybertill product to set the stock on.
 * @param int    $stockAmount The quantity of stock to set the product to have.
 */
	public function setStock($sku, $stockAmount) {
		if ($this->_setupClient()) {
			$productItem = $this->_getItem($sku);

			if ($productItem['success'] === true) {
				$newStock = [
					[
						'itemId' => $productItem['item']->productOption->id,
						'locationId' => $this->_config['location_id'],
						'reasonText' => 'Evoluted Add Stock',
						'qty' => $stockAmount,
						'updateType' => 'set'
					]
				];

				$this->_client->stock_update($newStock);
			}
		}
	}

/**
 * Create the soap client to create orders or sync stock. Uses settings from the config, if a requried
 * config setting is missing then the connection won't be made.
 */
	protected function _setupClient() {
		$this->_getConfig();

		if (!empty($this->_config['client_details'])) {
			$clientDetails = $this->_config['client_details'];
			$this->_client = new SoapClient($clientDetails['wsdl']);

			if (!empty($clientDetails['url']) && !empty($clientDetails['user_id'])) {
				$this->_authId = $this->_client->authenticate_get($clientDetails['url'], $clientDetails['user_id']);

				$this->_client = new SoapClient($clientDetails['wsdl'], ['login' => $this->_authId]);

				return true;
			}
		}

		if (Configure::check('EvCybertill.errorLogging.clientSetupFailure') && Configure::read('EvCybertill.errorLogging.clientSetupFailure') === true) {
			// A client hasn't been able to be made
			if (CakePlugin::loaded('EvErrbit')) {
				$e = new Exception("A client wasn't able to be established.");
				Errbit::notify($e);
			}
		}

		return false;
	}

/**
 * Read the config settings based on the current app environment. Allows for different testing settings if available.
 */
	protected function _getConfig() {
		if (Configure::read('app.environment') == 'PRODUCTION') {
			$this->_config = Configure::read('EvCybertill.live');
		} else {
			$this->_config = Configure::read('EvCybertill.dev');
		}
	}

/**
 * Attempt to get a customer using the provided customer id in the config. If an id is missing in the config
 * then attempt to create a customer using the customer details in the config. If a customer is found or
 * added then save the details
 */
	protected function _setupCustomer() {
		if (!empty($this->_config['customer_id'])) {
			$this->_customer = $this->_client->customer_get($this->_config['customer_id']);
		} else {
			if (!empty($this->_config['customer_details'])) {
				$this->_customer = $this->_client->customer_add($this->_config['customer_details']);
			}
		}
	}

/**
 * Get all the cybertill items from the provided order. Each item is checked if it exists and is available before being
 * added to the returned array. If an item is missing then the order items aren't returned and the cybertill
 * transaction won't take place.
 * @param  array  $order An EvCheckout order
 * @return array         The cybertill order items or null if an item can't be found
 */
	protected function _getItemsForOrder($order) {
		$paths = $this->_config['paths'];

		if (!empty($paths['items'])) {
			$items = Hash::get($order, $paths['items']);

			$itemErrorMessage = '';

			foreach ($items as $item) {
				$orderItem = $this->_getItemForOrder($item, $paths['sku']);
				if ($orderItem['success'] === true) {
					$orderItems[] = $orderItem['item'];
				} else {
					$itemErrorMessage .= "\n" . $orderItem['message'];
				}
			}

			if (!empty($orderItems) && count($orderItems) == count($items)) {
				return ['success' => true, 'data' => $orderItems];
			} else {
				return ['success' => false, 'message' => $itemErrorMessage];
			}
		}

		return ['success' => false, 'message' => 'Missing item path in config'];
	}

/**
 * Get an item for a cybertill order. Check if the item has stock available before trying to order it.
 * An cybertill order item array is constructed if the item is found and has stock and returned to add
 * to the order items array.
 * @param  array  $orderItem An item provided by EvCheckout to add to the cybertill order
 * @param  string $skuPath   The path within the orderItem array where the sku is stored.
 * @return array             The cybertill order item or null if no item was found.
 */
	protected function _getItemForOrder($orderItem, $skuPath) {
		// Get the product the client wants us to add, we need this for the product ID.
		$sku = Hash::get($orderItem, $skuPath);

		// Check the item is in stock before we try and add it to an order.
		$stock = $this->checkStock($sku);

		$stockMessage = '';
		if ($stock['success'] === false) {
			$stockMessage = $stock['message'];
		}

		$productItem = $this->_getItem($sku);

		if ($productItem['success'] === true) {
			$returnData = [
				'success' => true,
				'item' => [
					'itemId' => $productItem['item']->productOption->id,
					'salesQty' => $orderItem['quantity'],
					'note' => '',
					'vatRate' => $orderItem['tax_rate']
				]
			];

			// when available, use the original unit price. this will
			// be available in the array when a bulk discount has been used
			if (isset($orderItem['original_unit_price_inc_tax'])) {
				$returnData['item']['itemPrice'] = $orderItem['original_unit_price_inc_tax'];
			} else {
				$returnData['item']['itemPrice'] = $orderItem['unit_price_inc_tax'];
			}

			return $returnData;
		} else {
			return ['success' => false, 'message' => $productItem['message'] . ' ' . $stockMessage];
		}
	}

/**
 * Get an item from cybertill using the SKU provided.
 * @param  string $sku The SKU of the item to get from cybertill
 * @return object      The cybertill item or null if no item was found
 */
	protected function _getItem($sku) {
		try {
			return ['success' => true, 'item' => $this->_client->item_get(null, $sku, null, true)];
		} catch (Exception $e) {
			if (Configure::check('EvCybertill.errorLogging.stockNotFound') && Configure::read('EvCybertill.errorLogging.stockNotFound') === true) {
				if (CakePlugin::loaded('EvErrbit')) {
					$e = new Exception("no product exists with sku: " . $sku);
					Errbit::notify($e);
				}

				CakeLog::write('cybertillApi', "no product exists with sku: " . $sku);
				CakeLog::write('missingProduct', $sku);
			}

			return ['success' => false, 'message' => 'no product exists with sku: ' . $sku];
		}
	}

/**
 * Get the stock for a specific cybertill item. Items are found via their SKU's.
 * It is possible that if the item has been discountinued that an item would be found but no stock
 * exists for it. In this case then an item won't be returned
 * @param  string $sku        The SKU of the item to get stock of
 * @param  int    $itemId     The cybertill if of the item to get stock of
 * @return int                The current stock in cybertill of the item to get stock of
 */
	protected function _getItemStock($sku, $itemId) {
		$stock = null;
		if (!empty($itemId)) {
			try {
				$stock = $this->_client->stock_item($itemId, $this->_config['location_id']);
			} catch (Exception $e) {
				if (Configure::check('EvCybertill.errorLogging.stockNotFound') && Configure::read('EvCybertill.errorLogging.stockNotFound') === true) {
					if (CakePlugin::loaded('EvErrbit')) {
						$e = new Exception("no stock found for: " . $sku);
						Errbit::notify($e);
					}

					CakeLog::write('cybertillApi', "no stock found for: " . $sku);
					CakeLog::write('missingStock', $sku);
				}
			}
		}
		return $stock;
	}

/**
 * Get the order details for a cybertill order. The details are mainly taken from the config but the order total
 * is taken from the order created on the webite.
 * @param  array $order An array containing the order to get the order total from
 * @return array        The array containing the order details
 */
	protected function _getDetailsForOrder($order) {
		$orderDetails = $this->_config['order_details'];

		if (!empty($orderDetails)) {
			if (!empty($orderDetails['order_total_label'])) {
				//Get the order total from the order
				$orderDetails['order_total'] = $this->_getOrderTotalForOrder($order, $orderDetails['order_total_label']);

				if (!empty($orderDetails['order_total'])) {
					return [
						'success' => true,
						'data' => [
							'websiteId' => $orderDetails['website_id'],
							'customerId' => $this->_customer->customer->id,
							'locationId' => $this->_config['location_id'],
							'orderTotal' => $orderDetails['order_total'],
							'orderNote' => $orderDetails['order_note'],
							'customerOrderRef' => $orderDetails['customer_order_reference'],
						]
					];
				}
			}
		}
	}

/**
 * Get a total of an order to be used in the payment details and order details of the cybertill
 * order. The specific total to get is based on the label passed through which itself is in the config.
 * @param  array  $order An array containing the order to get the order total from
 * @param  string $label The lable of the total to return
 * @return float         The amount of total. If a tax inclusive price is available then it will return that.
 */
	protected function _getOrderTotalForOrder($order, $label) {
		$orderTotal = null;

		if (!empty($this->_config['paths']['totals'])) {
			$totals = Hash::get($order, $this->_config['paths']['totals']);

			foreach ($totals as $total) {
				if ($total['name'] == $label) {
					if (!empty($total['display_inc_tax'])) {
						$orderTotal = $total['display_inc_tax'];
					} else {
						$orderTotal = $total['amount'];
					}
				}
			}
		}

		return $orderTotal;
	}

/**
 * Get the delivery details for a cybertill order. The details are based on the user and as we use one user for all orders
 * we can't change it based on the actual delivery address of the order.
 * @param  array  $order An array containing the order to get the delivery details from
 * @return array         The array containing the payment details
 */
	protected function _getDeliveryForOrder($order) {
		return [
			'success' => true,
			'data' => [
				'isCollection' => true,
				'addressId' => $this->_customer->addresses->item->id,
				'serviceId' => null,
				'tariffId' => null,
				'tariff' => 0,
				'vatRate' => null,
				'recipient' => (!empty($this->_config['order_delivery']['recipient'])) ? $this->_config['order_delivery']['recipient'] : '',
				'dateRequired' => '',
				'when' => null,
				'instructions' => (!empty($this->_config['order_delivery']['instructions'])) ? $this->_config['order_delivery']['instructions'] : '',
				'giftMessage' => (!empty($this->_config['order_delivery']['giftMessage'])) ? $this->_config['order_delivery']['giftMessage'] : '',
				'giftReceipt' => false
			]
		];
	}

/**
 * Get the payment details for a cybertill order. By default the payment is recorded as a card payment unless
 * specified otherwise in the config.
 * @param  array  $order An array containing the order to get the payment details from
 * @return array         The array containing the payment details
 */
	protected function _getPaymentsForOrder($order) {
		$paymentDetails = $this->_config['order_payments'];

		if (!empty($paymentDetails['total_label'])) {
			//Get the payment total from the order
			$orderDetails['total'] = $this->_getOrderTotalForOrder($order, $paymentDetails['total_label']);

			return [
				'success' => true,
				'data' => [
					[
						'type' => (!empty($this->_config['order_payments']['type'])) ? $this->_config['order_payments']['type'] : 2,
						'total' => $orderDetails['total']
					]
				]
			];
		}
	}

/**
 * Update the stock in the website database using the stock from cybertill
 * @param  int 	  $cybertillStock The amount of stock currently in cybertill
 * @param  array  $item           The item to update
 * @param  obj    &$ItemModel     The Model of the item to update
 * @return bool                   Whether the update was successful or not
 */
	protected function _updateStock($cybertillStock, $item, &$ItemModel) {
		// Update the inventory with the stock from cybertill
		$StockModel = EvClassRegistry::init($this->_config['models']['stock_model']);

		$stock = $StockModel->find(
			'first',
			[
				'conditions' => [
					'id' => $item['Inventory']['id']
				]
			]
		);

		//Item exists on site so we can update inventory
		$stock[$StockModel->alias]['stock'] = $cybertillStock->item->stock;

		return $StockModel->save($stock);
	}
}
